/* HTTP client requests
 * (mostly GET)
 *
 * Copyright (c) 2004 David Hammerton
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

/* PRIVATE */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netdb.h>

#include "config.h"

#include "http_client.h"
#include "debug/debug.h"

#define DEFAULT_DEBUG_CHANNEL "http_client"

struct HTTP_ConnectionTAG
{
    char *host;
    int sockfd;
};

typedef struct HTTP_HeaderTAG HTTP_Header;
struct HTTP_HeaderTAG
{
    char *field_name;
    char *field_value;
    HTTP_Header *next;
};

static int HTTP_Connect(const char *host, const char *port)
{
    int sockfd, ret;
    struct addrinfo hints, *res, *ressave;

    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    ret = getaddrinfo(host, port, &hints, &res);

    if (ret)
    {
        ERR("getaddrinfo error: [%s] (%s)\n",
            gai_strerror(ret), host);
        return -1;
    }

    ressave = res;
    sockfd = -1;

    while (res)
    {
        sockfd = socket(res->ai_family,
                        res->ai_socktype,
                        res->ai_protocol);
        if (sockfd >= 0)
        {
            if (connect(sockfd, res->ai_addr, res->ai_addrlen) == 0)
                break;

            close (sockfd);
            sockfd = -1;
        }
        res = res->ai_next;
    }

    freeaddrinfo(ressave);

    return sockfd;
}

static void HTTP_Close(int sockfd)
{
    close(sockfd);
}

HTTP_Connection *HTTP_Client_Open(const char *host)
{
    HTTP_Connection *new;
    int sockfd = HTTP_Connect(host, "3689");
    if (sockfd == -1)
        return NULL;
    new = malloc(sizeof(HTTP_Connection));
    new->sockfd = sockfd;
    new->host = malloc(strlen(host)+1);
    strcpy(new->host, host);
    return new;
}

static void HTTP_Client_Reset(HTTP_Connection *connection)
{
    HTTP_Close(connection->sockfd);
    connection->sockfd = HTTP_Connect(connection->host, "3689");
    if (connection->sockfd == -1)
        ERR("unhandled error\n");
}

void HTTP_Client_Close(HTTP_Connection *connection)
{
    HTTP_Close(connection->sockfd);
    free(connection->host);
    free(connection);
}

#define IS_CRLF(p) ((*p) == '\n' || (*p) == '\r')

static int HTTP_ProcessHeaders(char *curin, int curlen,
                               char **curout,
                               int *requiredExtraLen,
                               char **contentout, int *contentoutlen,
                               HTTP_Header **headers)
{
    char *curHeaderStart = curin;
    int curHeaderLen = 0;
    int n = 0;

    *requiredExtraLen = 0;
    *contentout = NULL;

    while (n < curlen)
    {
        HTTP_Header *header;
        int i;
        /* FIXME: there is a bug here. if the previous header ends on a single
         * \r, then the next header this will trigger with the \n and we'll
         * quit early :(
         */
        if (IS_CRLF(curHeaderStart))
        { /* end of header */
            char *content = curHeaderStart + 1;

            if (n < curlen-1)
            {
                if (IS_CRLF(content))
                {
                    content++;
                    n++;
                }
                *contentout = content;
                *contentoutlen = curlen - n - 1;
            }
            return 0;
        }
        while (!IS_CRLF(&curHeaderStart[curHeaderLen]))
        {
            curHeaderLen++;
            n++;
            if (n >= curlen)
            {
                if (curHeaderStart == curin)
                    *requiredExtraLen = 1;
                *curout = curHeaderStart;
                return 1;
            }
        }
        n++;
        curHeaderLen++;
        if (IS_CRLF(&curin[n]))
            n++;

        /* sizeof(HTTP_Header) + curHeaderLen + \0 + \0 - ':' - ' '
         * + (possibly a \0 if its not a real header) = 1 */
        header = malloc(sizeof(HTTP_Header) + curHeaderLen + 1);

        i = 0;
        header->field_name = (char*)header + sizeof(HTTP_Header);
        while (curHeaderStart[i] != ':' && i < curHeaderLen)
        {
            header->field_name[i] = curHeaderStart[i];
            i++;
        }

        if (curHeaderStart[i] == ':')
        {
            header->field_name[i] = 0;
            header->field_value = (char*)header + sizeof(HTTP_Header) +
                                  strlen(header->field_name) + 1;
            strncpy(header->field_value, &curHeaderStart[i+2],
                    curHeaderLen - i - 2);
            header->field_value[curHeaderLen - i - 3] = 0;
        }
        else
        {
            header->field_name[curHeaderLen-1] = 0;
            header->field_value = NULL;
        }
        header->next = NULL;

        /* we want to store it at the tail, but we are given the head */
        if (!(*headers))
        {
            *headers = header;
        }
        else
        {
            HTTP_Header *curIter = *headers;
            while (curIter->next)
            {
                curIter = curIter->next;
            }
            curIter->next = header;
        }

        curHeaderStart = &curin[n];
        curHeaderLen = 0;
    }
    *curout = NULL;
    return 1;
}

int HTTP_PassStandardHeaders(HTTP_Header *headersList,
                             int *httpContentLength)
{
    int httpStatusCode;
    HTTP_Header *cur;

    *httpContentLength = 0;

    /* first get the first header, this will be status code */
    if (headersList->field_value == NULL &&
            strncmp(headersList->field_name, "HTTP/1.1 ", 8) == 0)
    {
        TRACE("http status line: '%s'\n", headersList->field_name);
        httpStatusCode = atoi(headersList->field_name + 8);
    }
    else
    {
        ERR("HTTP status code wan't first\n");
        return 0;
    }

    /* now find content length (then return) */
    cur = headersList;
    while (cur)
    {
        if (strcmp(cur->field_name, "Content-Length") == 0)
        {
            *httpContentLength = atoi(cur->field_value);
            return httpStatusCode;
        }
        cur = cur->next;
    }

    return httpStatusCode;
}

#define METHOD_GET "GET "
#define HTTP_VERSION " HTTP/1.1\r\n"
#define HEADER_HOST "Host: "
#define DAAP_VERSION "\r\nClient-DAAP-Version: 3.0\r\n"
#define HEADER_USERAGENT "User-Agent: iTunes/4.5 (Windows; N)\r\n"
#define HEADER_ACCEPT "Accept-Language: en-us, en;q=5.0\r\n"
#define DAAP_VALIDATE "Client-DAAP-Validation: "
#define DAAP_ACCESS_INDEX "Client-DAAP-Access-Index: 2\r\n"
#define ACCEPT_ENCODING "Accept-Encoding: gzip\r\n"
#define CONNECTION_CLOSE "Connection: close\r\n"
#define END_REQUEST "\r\n"

static int HTTP_Client_RequestGet(const HTTP_Connection *connection,
                                  const char *path, const char *hash,
                                  const char *extra_header,
                                  int send_close)
{
    int buffer_size = 0;
    char *buffer = NULL;
    char *buffersend = NULL;
    int tosend;
    int ret, res = 1;

    buffer_size += strlen(METHOD_GET);
    buffer_size += strlen(path);
    buffer_size += strlen(HTTP_VERSION);
    buffer_size += strlen(HEADER_HOST);
    buffer_size += strlen(connection->host);
    buffer_size += strlen(DAAP_VERSION);
    buffer_size += strlen(HEADER_USERAGENT);
    buffer_size += strlen(HEADER_ACCEPT);
    if (send_close)
        buffer_size += strlen(CONNECTION_CLOSE);
    buffer_size += strlen(DAAP_ACCESS_INDEX);
    if (hash)
    {
        buffer_size += strlen(DAAP_VALIDATE);
        buffer_size += strlen(hash);
        buffer_size += strlen(END_REQUEST);
    }
//    buffer_size += strlen(ACCEPT_ENCODING);
    buffer_size += strlen(END_REQUEST);
    if (extra_header) buffer_size += strlen(extra_header);
    buffer_size += 1;

    buffer = malloc(buffer_size);
    buffer[0] = 0;
    strcat(buffer, METHOD_GET);
    strcat(buffer, path);
    strcat(buffer, HTTP_VERSION);
    strcat(buffer, HEADER_HOST);
    strcat(buffer, connection->host);
    strcat(buffer, DAAP_VERSION);
    strcat(buffer, HEADER_USERAGENT);
    strcat(buffer, HEADER_ACCEPT);
    if (send_close)
        strcat(buffer, CONNECTION_CLOSE);
    strcat(buffer, DAAP_ACCESS_INDEX);
    if (hash)
    {
        strcat(buffer, DAAP_VALIDATE);
        strcat(buffer, hash);
        strcat(buffer, END_REQUEST);
    }
//    strcat(buffer, ACCEPT_ENCODING);
    if (extra_header) strcat(buffer, extra_header);
    strcat(buffer, END_REQUEST);
    /* just for printing. we shouldn't send this; */
    buffer[buffer_size-1] = 0;

#if 0
    TRACE("about to send something of size %i\n", buffer_size);
    DPRINTF("%s\n\n", buffer);
    debug_hexdump(buffer, buffer_size);
#endif

    buffersend = buffer;
    tosend = buffer_size - 1;
    while (tosend)
    {
        ret = send(connection->sockfd, buffersend, tosend, 0);
        if (ret == -1)
        {
            ERR("send error: [%s]\n", strerror(errno));
            res = 0;
            goto end;
        }
        tosend -= ret;
        buffersend -= ret;
    }

end:
    free(buffer);
    return res;
}

static char *HTTP_Client_ReadHeaders(const HTTP_Connection *connection,
                                     HTTP_Header **headersList,
                                     char **contentFromHeaders,
                                     int *contentLenFromHeaders)
{
    char *buffer = NULL;
    int buffer_size;
    int ret;
    char *headers;
    int headers_needmore = 0;

    buffer_size = 1000;
    buffer = malloc(buffer_size);
    headers = NULL;
    do
    {
        int read_max = buffer_size;
        if (headers)
        {
            char *new;
            int required_copy;

            required_copy = buffer_size - (headers - buffer);
            if (headers_needmore)
                buffer_size *= 2;
            new = malloc(buffer_size);
            memcpy(new, headers, required_copy);

            free(buffer);
            buffer = new;

            headers = buffer + required_copy;
            read_max = buffer_size - required_copy;
        }
        else headers = buffer;
        ret = recv(connection->sockfd, headers, read_max, 0);
        if (ret == -1)
        {
            ERR("an error occured on recv!\n");
            free(buffer);
            goto end;
        }
        //debug_hexdump(buffer, ret);
    }
    while (HTTP_ProcessHeaders(buffer, ret,
                               &headers,
                               &headers_needmore,
                               contentFromHeaders, contentLenFromHeaders,
                               headersList));

    if (!*headersList)
        goto end;

    return buffer;
end:
    free(buffer);
    return NULL;

}

HTTP_GetResult *HTTP_Client_Get(const HTTP_Connection *connection,
                                const char *path,
                                const char *hash,
                                const char *extra_header,
                                int reset_send_close)
{
    char *databuffer = NULL;
    int torecv;
    int ret;
    char *headerbuffer;

    char *contentFromHeaders;
    int contentLenFromHeaders;
    HTTP_Header *headersList = NULL;

    HTTP_GetResult *result;

    int httpStatusCode;
    int httpContentLength;

    if (!HTTP_Client_RequestGet(connection, path, hash,
                                extra_header, reset_send_close))
        goto end;

    if (!(headerbuffer = HTTP_Client_ReadHeaders(connection, &headersList,
                                 &contentFromHeaders, &contentLenFromHeaders)))
    {
        ERR("failed to recieve any headers\n");
        goto end;
    }

    httpStatusCode = HTTP_PassStandardHeaders(headersList,
                                              &httpContentLength);

    if (httpStatusCode != 200)
    {
        ERR("invalid status code [%i]\n", httpStatusCode);
        goto end;
    }
    if (httpContentLength == 0)
    {
        ERR("no content length\n");
        goto end;
    }

    result = malloc(sizeof(HTTP_GetResult) + httpContentLength);
    result->data = (char*)result + sizeof(HTTP_GetResult);
    result->contentlen = httpContentLength;

    torecv = httpContentLength;
    databuffer = result->data;
    if (contentFromHeaders)
    {
        memcpy(result->data, contentFromHeaders, contentLenFromHeaders);
        torecv -= contentLenFromHeaders;
        databuffer += contentLenFromHeaders;
    }
    free(headerbuffer);
    while (torecv)
    {
        ret = recv(connection->sockfd, databuffer, torecv, 0);
        if (ret == -1)
        {
            ERR("an error occured on recv\n");
            goto end;
        }
        torecv -= ret;
        databuffer += ret;
    }


    if (reset_send_close)
        HTTP_Client_Reset(connection);

    return result;
end:
    if (reset_send_close)
        HTTP_Client_Reset(connection);
    ERR("returning with error\n");

    return NULL;
}

/* used for async io */
int HTTP_Client_Get_ToFile(HTTP_Connection *connection,
                           const char *path,
                           const char *hash,
                           const char *extra_header,
                           int filed,
                           int (*pfnStatus)(void*,int),
                           void *userdata,
                           int reset_send_close)
{
    char *headerbuffer;
    char *contentFromHeaders;
    int contentLenFromHeaders;
    HTTP_Header *headersList = NULL;

    int httpStatusCode;
    int httpContentLength;

    char readbuffer[1024];
    int torecv;

    int last_micropercent = 0;
    int micropercent;


    if (!HTTP_Client_RequestGet(connection, path, hash,
                                extra_header, reset_send_close))
        goto end;

    if (!(headerbuffer = HTTP_Client_ReadHeaders(connection, &headersList,
                                 &contentFromHeaders, &contentLenFromHeaders)))
    {
        ERR("failed to recieve any headers\n");
        goto end;
    }

    httpStatusCode = HTTP_PassStandardHeaders(headersList,
                                              &httpContentLength);

    if (httpStatusCode != 200)
    {
        ERR("invalid status code [%i]\n", httpStatusCode);
        goto end;
    }
    if (httpContentLength == 0)
    {
        ERR("no content length\n");
        goto end;
    }

    torecv = httpContentLength;
    /* FIXME: write in portions */
    if (contentFromHeaders)
    {
        write(filed, contentFromHeaders, contentLenFromHeaders);
        torecv -= contentLenFromHeaders;
    }
    free(headerbuffer);

    micropercent = ((float)(httpContentLength - torecv) /
                    (float)httpContentLength) * 1000.0;
    if (micropercent > last_micropercent)
    {
        if (pfnStatus(userdata, micropercent))
        {
            HTTP_Client_Reset(connection);
            goto end;
        }
        last_micropercent = micropercent;
    }


    while (torecv)
    {
        int ret;
        int readsize = (int)sizeof(readbuffer) < torecv ?
            (int)sizeof(readbuffer) : torecv;
        ret = recv(connection->sockfd, readbuffer, readsize, 0);
        if (ret == -1)
        {
            ERR("an error occured on recv\n");
            goto end;
        }
        torecv -= ret;
        write(filed, readbuffer, ret);
        micropercent = ((float)(httpContentLength - torecv) /
                         (float)httpContentLength) * 1000.0;
        if (micropercent > last_micropercent)
        {
            if (pfnStatus(userdata, micropercent))
            {
                HTTP_Client_Reset(connection);
                goto end;
            }
            last_micropercent = micropercent;
        }
    }

    if (reset_send_close)
        HTTP_Client_Reset(connection);
    return 1;
end:
    if (reset_send_close)
        HTTP_Client_Reset(connection);
    return 0;
}

void HTTP_Client_FreeResult(HTTP_GetResult *res)
{
    free(res);
}


